import requests
from typing import Dict, Any, List

"""
Request Timeout status code.
Technically not proper as the server should return the 408 in tradititional usage,
but it is a good enough indicator for our purposes
"""
REQUEST_TIMEOUT_STATUS_CODE = 408
DEFAULT_TIMEOUT = 10


def create_json(metadata, value=None):
    result = {"causal-metadata": {}}
    if metadata is not None:
        result["causal-metadata"] = metadata
    if value is not None:
        result["value"] = value
    return result


# client for kvs api


class KVSClient:
    def __init__(self, base_url: str):
        # set base url without trailing slash
        self.base_url = base_url if not base_url.endswith("/") else base_url[:-1]

    def ping(self, timeout: float = DEFAULT_TIMEOUT) -> requests.Response:
        if timeout is not None:
            try:
                return requests.get(f"{self.base_url}/ping", timeout=timeout)
            except requests.exceptions.Timeout:
                r = requests.Response()
                r.status_code = REQUEST_TIMEOUT_STATUS_CODE
                return r
        else:
            return requests.get(f"{self.base_url}/ping")

    def get(
        self, key: str, metadata: str, timeout: float = DEFAULT_TIMEOUT
    ) -> requests.Response:
        if not key:
            raise ValueError("key cannot be empty")

        if timeout is not None:
            try:
                return requests.get(
                    f"{self.base_url}/data/{key}",
                    json=create_json(metadata),
                    timeout=timeout,
                )
            except requests.exceptions.Timeout:
                r = requests.Response()
                r.status_code = REQUEST_TIMEOUT_STATUS_CODE
                return r
        else:
            return requests.get(
                f"{self.base_url}/data/{key}", json=create_json(metadata)
            )

    def put(
        self, key: str, value: str, metadata: str, timeout: float = DEFAULT_TIMEOUT
    ) -> requests.Response:
        if not key:
            raise ValueError("key cannot be empty")

        if timeout is not None:
            try:
                return requests.put(
                    f"{self.base_url}/data/{key}",
                    json=create_json(metadata, value),
                    timeout=timeout,
                )
            except requests.exceptions.Timeout:
                r = requests.Response()
                r.status_code = REQUEST_TIMEOUT_STATUS_CODE
                return r
        else:
            return requests.put(
                f"{self.base_url}/data/{key}", json=create_json(metadata, value)
            )

    def delete(self, key: str, timeout: float = DEFAULT_TIMEOUT) -> requests.Response:
        if not key:
            raise ValueError("key cannot be empty")

        if timeout is not None:
            try:
                return requests.delete(f"{self.base_url}/data/{key}", timeout=timeout)
            except requests.exceptions.Timeout:
                r = requests.Response()
                r.status_code = REQUEST_TIMEOUT_STATUS_CODE
                return r
        else:
            return requests.delete(f"{self.base_url}/data/{key}")

    def get_all(
        self, metadata: str, timeout: float = DEFAULT_TIMEOUT
    ) -> requests.Response:
        if timeout is not None:
            try:
                return requests.get(
                    f"{self.base_url}/data", json=create_json(metadata), timeout=timeout
                )
            except requests.exceptions.Timeout:
                r = requests.Response()
                r.status_code = REQUEST_TIMEOUT_STATUS_CODE
                return r
        else:
            return requests.get(f"{self.base_url}/data")

    def clear(self, timeout: float = DEFAULT_TIMEOUT) -> None:
        response = self.get_all(timeout=timeout)
        if response.status_code != 200:
            raise RuntimeError(f"failed to get keys: {response.status_code}")

        for key in response.json().keys():
            delete_response = self.delete(key, timeout=timeout)
            if delete_response.status_code != 200:
                raise RuntimeError(
                    f"failed to delete key {key}: {delete_response.status_code}"
                )

    def send_view(
        self, view: dict[str, List[Dict[str, Any]]], timeout: float = DEFAULT_TIMEOUT
    ) -> requests.Response:
        if not isinstance(view, dict):
            raise ValueError("view must be a dict")

        request_body = {"view": view}
        return requests.put(f"{self.base_url}/view", json=request_body, timeout=timeout)
